/*    This file is part of Zmanim.

    Zmanim is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Zmanim is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zmanim; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef GUARD_KEN_OPERATORS_H
#define GUARD_KEN_OPERATORS_H
#include "config.h"

namespace kenOps{

	////////////////////////////////////////////////////////////////////////////
	// Comparable
	// use this instead of std::rel_ops
	
	template <class T, class U> struct comparable2
	{
		friend bool operator<=(const T& x, const U& y) { return !(x > y); }
		friend bool operator>=(const T& x, const U& y) { return !(x < y); }
		friend bool operator<(const U& x, const T& y)  { return y > x; }
		friend bool operator<=(const U& x, const T& y) { return !(y < x); }
		friend bool operator>=(const U& x, const T& y) { return !(y > x); }
		friend bool operator==(const U& y, const T& x) { return x == y; }
		friend bool operator!=(const U& y, const T& x) { return !(x == y); }
		friend bool operator!=(const T& y, const U& x) { return !(y == x); }
		
		// you must define the following
		//friend bool operator<(const T& x, const U& y);
		
	};
	
	template <class T> struct comparable1
	{
		friend bool operator>(const T& x, const T& y)  { return y < x; }
		friend bool operator<=(const T& x, const T& y) { return !(y < x); }
		friend bool operator>=(const T& x, const T& y) { return !(x < y); }
		friend bool operator!=(const T& x, const T& y) { return !(x == y); }
		
		// you must define the following
		//friend bool operator<(const T& x, const T& y);
		
	};
	
	////////////////////////////////////////////////////////////////////////////
	// Mathematical operators
	// this may be used only where the operators defined used are commutative.
	// to enable an operator (or set of operators), define the appropriate
	// op= operator.
	template <typename T, typename U> struct mathematics2
	{
		friend T operator+(T x, const U& y) { return x += y; }
		friend T operator+(const U& y, T x) { return x += y; }
		friend T operator-(T x, const U& y) { return x -= y; }
		friend T operator-(const U& y, T x) { return x -= y; }
		friend T operator*(T x, const U& y) { return x *= y; }
		friend T operator*(const U& y, T x) { return x *= y; }
		friend T operator/(T x, const U& y) { return x /= y; }
		friend T operator/(const U& y, T x) { return x /= y; }
		friend T operator%(T x, const U& y) { return x %= y; }
		friend T operator%(const U& y, T x) { return x %= y; }
	};

	template <typename T> struct mathematics1
	{
		friend T operator+(T x, const T& y) { return x += y; }
		friend T operator-(T x, const T& y) { return x -= y; }
		friend T operator*(T x, const T& y) { return x *= y; }
		friend T operator/(T x, const T& y) { return x /= y; }
		friend T operator%(T x, const T& y) { return x %= y; }
		// you must define the following
		//T& operator%= (const T& y);
	};

	////////////////////////////////////////////////////////////////////////////
	// Distance Operators
	// the operators here are not commutative.
	// this pattern is defined to appropriately handle patterns where there is
	// an absolute scale (such as date or time) and a different type for distances
	// between two absolutes
	template <class absolute_type, class distance_type>
		struct dist_operators
	{
		friend absolute_type operator+ (absolute_type x, distance_type y)
		{ return x += y ;}
		friend absolute_type operator- (absolute_type x, distance_type y)
		{ return x -= y ;}

		// to use this pattern, you must define:
		// absolute_type& absolute_type::operator+= (distance_type y);
		// absolute_type& absolute_type::operator-= (distance_type y);
		// distance_type operator-(absolute_type x, absolute_type, y);
	};

	////////////////////////////////////////////////////////////////////////////
	// Increments and decrements
	// insures appropriate postincrement and postdecrement semantics
	template <class T> struct post_inc_dec
	{
		friend T operator++ (T& orig, int)
		{
			T retval=orig;
			++orig;
			return retval;
		}

		friend T operator-- (T& orig, int)
		{
			T retval=orig;
			--orig;
			return retval;
		}
	};
}

#endif

