/*    This file is part of Zmanim.

    Zmanim is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Zmanim is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zmanim; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//sunfuncs.cpp
//Ken Bloom

#include "config.h"
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "date.h"
#include "sunfuncs.h"

using namespace kenDate;

double const pi = acos(-1);

inline double toRad(double degrees)
{
	return degrees * pi / 180;
}

inline double toRad(int degrees, int minutes)
{
	return toRad (double (degrees) + double (minutes) / 60);
}

inline double toDeg(double rad)
{
	return rad * 180 / pi;
}

double sun_rise_set(date theday,
               double latitude, double longitude, double zenith, bool sunset)
//returns time in UTC
{

/*
Based on Almanac for Computers, US Naval Observatory

see also: http://williams.best.vwh.net/sunrise_sunset_algorithm.htm
*/

	//step 1: First calculate the day of the year
	int N=theday - date(1,1,theday.year()) + 1;

	//step 2: convert the longitude to hour value and calculate an approximate time
	double lngHour= longitude / 15;

	double t = N + ((6 - lngHour) / 24);
	if (sunset) t= N + ((18 - lngHour) / 24);

	//step 3: calculate the sun's mean anomaly
	double M = (0.9856 * t) - 3.289;

	//step 4: calculate the sun's true longitude
	double L = M + (1.916 * sin(toRad(M))) + (0.020 * sin (toRad(2 * M))) + 282.634;
	while (L<0) L+=360;
	while (L>=360) L-=360;

	//step 5a: calculate the sun's right ascension
	double RA = toDeg(atan (0.91764 * tan(toRad(L))));
	while (RA<0)RA+=360;
	while (RA>=360)RA-=360;
	
	//step 5b: right ascension value needs to be in the same quadrant as L
	double Lquadrant = floor(L/90)*90;
	double RAquadrant = floor(RA/90)*90;
	RA=RA+(Lquadrant - RAquadrant);

	//step 5c: right ascension value needs to be converted into hours
	RA/=15;

	//step 6: calculate the sun's declination
	double sinDec = 0.39782 * sin(toRad(L));
	double cosDec = cos(asin(sinDec));

	//step 7a: calculate the sun's local hour angle
	double cosH = (cos(toRad(zenith)) - (sinDec * sin(toRad(latitude)))) / (cosDec * cos(toRad(latitude)));

	if (cosH > 1 || cosH < -1)
		throw doesnthappen();

	//step 7b: finish calculating H and convert into hours
	double H = 360 - toDeg(acos(cosH));
	if (sunset) H=toDeg(acos(cosH));

	H=H/15;

	//step 8: calculate local mean time
	
	double T = H + RA - (0.06571 * t) - 6.622;
	while(T >= 24) T-=24;
	while(T<0) T+=24;

	//step 9: convert to UTC. numbers outside the range 0..24 
	//indicate the previous day UTC or the next day UTC
	double UT = T - lngHour;

	return UT;
}

double sunrise(date theday,
               double latitude, double longitude, double zenith)
{
   return sun_rise_set(theday,latitude,longitude,zenith,false);
}
double sunset(date theday,
               double latitude, double longitude, double zenith)
{
   return sun_rise_set(theday,latitude,longitude,zenith,true);
}

date convert_to_tz(date theday, double thetime, std::string timezone){
   struct tm tm;
   time_t timenum;

   //mktime has the side effect of normalizing tm
   tm.tm_min=tm.tm_hour=0;
   tm.tm_sec=static_cast<int>(thetime*60*60);
   tm.tm_mon=theday.month()-1;
   tm.tm_mday=theday.day();
   tm.tm_year=theday.year()-1900;
   tm.tm_isdst=0;
   //tm.tm_gmtoff=0;
   

   //Get timenum for the UTC time.
   putenv("TZ=UTC");
   tzset();
   timenum=mktime(&tm);

   //Convert back to the requested local timezone
#if defined(HAVE_SETENV)
   setenv("TZ",timezone.c_str(),1);
#elif defined(HAVE_PUTENV)
   char* newTZbuf=(char*) malloc(timezone.length() * sizeof(char) + 5);
   sprintf(newTZbuf,"TZ=%s",timezone.c_str());
   putenv(newTZbuf);
#else
   #error "Your system lacks both putenv() and setenv(), one of which is required to compile this program"
#endif
   tzset();
   localtime_r(&timenum,&tm);
   
   //return a value
   date returnDate(tm.tm_mon+1,tm.tm_mday,tm.tm_year+1900);
   returnDate.hour(tm.tm_hour);
   returnDate.min(tm.tm_min);
   return returnDate;
}
